from __future__ import print_function

import os
import re
import subprocess
import sys
import tempfile


class BibtexException(Exception):
    def __init__(self, msg, use_verbose=False):
        super(BibtexException, self).__init__(msg)
        self.use_verbose = use_verbose


def cite(keys, database, bibstyle="plain", verbose=False):
    if not keys:
        return {}
    aux_filepath = _write_aux_file(keys, database, bibstyle)
    _bibtex(aux_filepath, verbose)
    bbl_filepath = aux_filepath.replace(".aux", ".bbl")
    return _parse_bbl(bbl_filepath)


def fallback(entry):
    fields = entry["fields"]
    if "author" in fields and "year" in fields and "title" in fields:
        return "{author} ({year}). {title}".format(**fields)
    if "author" in fields and "year" in fields:
        return "{author} ({year})".format(**fields)
    if "author" in fields and "title" in fields:
        return "{author}. {title}".format(**fields)
    if "title" in fields and "year" in fields:
        return "{title} ({year})".format(**fields)
    return entry["type"]


def _write_aux_file(keys, database, bibstyle):
    auxfile = tempfile.NamedTemporaryFile("w", suffix=".aux", delete=False)
    with auxfile as f:
        print(r"\bibdata{{{}}}".format(database), file=f)
        print(r"\bibstyle{{{}}}".format(bibstyle), file=f)
        for key in keys:
            line = r"\citation{{{}}}".format(key)
            print(line, file=f)

    return auxfile.name


def _bibtex(aux_filepath, verbose):
    cwd, aux_filename = os.path.split(aux_filepath)
    stdout = sys.stdout if verbose else subprocess.PIPE
    stderr = sys.stderr if verbose else subprocess.PIPE
    try:
        p = subprocess.Popen(
            ["bibtex", aux_filename], cwd=cwd, stdout=stdout, stderr=stderr
        )
    except OSError:  # Common for py 2 and 3 and parent of FileNotFoundError
        raise BibtexException("bibtex is not available")
    p.wait()
    if p.returncode != 0:
        raise BibtexException("bibtex failed", use_verbose=True)


bibitem_pattern = re.compile(r"^\\bibitem(\[.*\])?\{(?P<key>.*)\}\n")


def _parse_bbl(bbl_filepath):
    with open(bbl_filepath) as f:
        content = f.read()

    to_return = {}

    raw_items = content.split("\n\n")

    for raw in raw_items[1:-1]:
        key = bibitem_pattern.search(raw).group("key")
        to_return[key] = _process_text(raw)

    return to_return


def _process_text(text):
    text = re.sub(bibitem_pattern, "", text)
    text = " ".join(l.strip() for l in text.splitlines())
    lines = text.split("\\newblock ")
    text = " ".join(_process_line(l) for l in lines)
    return text


def _process_line(line):
    line = line.strip()
    line = line.replace("~", " ")
    line = re.sub(r"\{\\em (.*)\}", r"\1", line)
    return line



"""
Command line reference manager with a single source of truth: the .bib file.
Inspired by beets.
"""

import click
import click_constraints
import click_plugins  # type: ignore
import pybibs
import pylatexenc.latex2text  # type: ignore
import pyperclip  # type: ignore
import requests

from . import cite
from . import internals
from . import query

FILE_OPTIONS = internals.combine_decorators(
    [
        click.option(
            "--file",
            help="Path to file to link to this entry.",
            type=click.Path(exists=True, readable=True, dir_okay=False),
        ),
        click.option(
            "--destination",
            help="A folder to put the file in.",
            type=click.Path(exists=True, readable=True, dir_okay=True, file_okay=False),
        ),
        click.option(
            "--no-copy",
            help="Add the specified file in its current location without copying.",
            is_flag=True,
        ),
        click_constraints.constrain("destination", depends=["file"]),
        click_constraints.constrain(
            "no_copy",
            depends=["file"],
            conflicts=["destination"],
        ),
    ]
)
SEARCH_TERMS_OPTION = click.argument(
    "search_term",
    nargs=-1,
    shell_complete=internals.complete_key,
)


@click_plugins.with_plugins(internals.get_plugins())
@click.group(help=__doc__)
@click.version_option()
@click.option(
    "--database",
    envvar=internals.BIBO_DATABASE_ENV_VAR,
    help="""
A path to a .bib file. Overrides the BIBO_DATABASE environment variable.
""",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
)
@click.pass_context
def cli(ctx, database):
    ctx.ensure_object(dict)
    ctx.obj["database"] = database
    ctx.obj["data"] = internals.load_database(database)


@cli.command("list", short_help="List entries.")
@click.option("--raw", is_flag=True, help="Format as raw .bib entries.")
@click.option(
    "--bibstyle",
    default="plain",
    help="""
Bibtex bibliography style to use for citation formatting.
For more information check https://www.overleaf.com/learn/latex/Bibtex_bibliography_styles.
""",
)
@click.option(
    "--format",
    help="""
Custom format pattern.
Use ``$`` in front of a key, type, or field to create custom formatter.
For example: ``--format '$author ($year) - $title'``.
""",
)
@click.option("--verbose", is_flag=True, help="Show verbose information.")
@SEARCH_TERMS_OPTION
@click.pass_context
def list_(ctx, search_term, raw, bibstyle, verbose, **kwargs):
    """
    List entries in the database.

    A SEARCH_TERM matches an entry if it appears in the type, key, or any
    of the fields of the entry.
    If multiple search terms are provided an entry should match all of them.
    It is possible to match against a specific key, type, or field as
    follows: ``author:einstein``, ``year:2018`` or ``type:book``.
    Note that search terms are case insensitive.
    """

    format_pattern = kwargs.pop("format")
    assert not kwargs

    results = query.search(ctx.obj["data"], search_term)
    if raw:
        _list_raw((r.entry for r in results))
    elif format_pattern:
        _list_format_pattern((r.entry for r in results), format_pattern)
    else:
        _list_citations(results, ctx.obj["database"], bibstyle, verbose)


def _list_raw(entries):
    for entry in entries:
        click.echo(pybibs.write_string([entry]))


def _list_format_pattern(entries, format_pattern):
    for entry in entries:
        click.echo(internals.format_entry(entry, format_pattern))


def _list_citations(results, database, bibstyle, verbose):
    results = list(results)
    keys = [r.entry["key"] for r in results]
    exception = None
    try:
        citations = cite.cite(keys, database, bibstyle, verbose)
    except cite.BibtexException as e:
        exception = e

    latex_converter = pylatexenc.latex2text.LatexNodes2Text()

    for result in results:
        header = internals.header(result.entry)
        if exception:
            citation = cite.fallback(result.entry)
        else:
            citation = citations[result.entry["key"]]

        citation = latex_converter.latex_to_text(citation)
        text = "\n".join([header, citation])
        text, extra_match_info = internals.highlight_match(text, result)

        click.echo(text)

        if extra_match_info:
            click.secho("Search matched by", underline=True)
        for key, val in extra_match_info.items():
            click.echo(f"{key}: {latex_converter.latex_to_text(val)}")

    if exception is not None:
        parts = [str(exception), "Using a fallback citation method"]
        if exception.use_verbose:
            parts.append("Use --verbose for more information")
        click.secho(". ".join(parts), fg="red")


@cli.command("open", short_help="Open the file, URL, or doi associated with an entry.")
@SEARCH_TERMS_OPTION
@click.pass_context
def open_(ctx, search_term):
    """
    Open an entry in the database if a ``file``, ``url``, or ``doi`` field
    exists (with precedence in this order).

    A file will be open by the application defined by your system according
    to the file extension.
    For example, a PDF should be opened by a PDF reader and a folder should
    be opened by a file browser.
    URLs and DOIs should be opened in the web browser

    A SEARCH_TERM matches an entry if it appears in the type, key, or any
    of the fields of the entry.
    If multiple search terms are provided an entry should match all of them.
    It is possible to match against a specific key, type, or field as
    follows: ``author:einstein``, ``year:2018`` or ``type:book``.
    Note that search terms are case insensitive.

    This command fails if the number of entries that match the search is
    different than one.
    """
    entry = query.get(ctx.obj["data"], search_term).entry

    for field_name in ["file", "url", "doi"]:
        value = entry.get("fields", {}).get(field_name)
        if value:
            if field_name == "doi":
                value = "https://doi.org/" + value
            click.launch(value)
            break
    else:
        raise click.ClickException("No file, url, or doi is associated with this entry")


@cli.command(short_help="Add a new entry.")
@FILE_OPTIONS
@click.option("--doi", help="Add entry by DOI.")
@click.pass_context
def add(ctx, destination, doi, no_copy, **kwargs):
    """
    Add a new entry to the database.

    Find a bib entry you would like to add.
    Copy it to the clipboard, and run the command.
    It will be opened in your editor for validation or manual editing.
    Upon saving, the entry is added to the database.

    Don't forget to set the EDITOR environment variable for this command
    to work properly.
    """
    file_ = kwargs.pop("file")

    data = ctx.obj["data"]
    if doi is not None:
        url = "http://dx.doi.org/{}".format(doi)
        headers = {"Accept": "application/x-bibtex"}
        resp = requests.get(url, headers=headers)
        assert resp.status_code == 200
        raw_bib = resp.text
    else:
        raw_bib = pyperclip.paste()
    bib = internals.editor(text=raw_bib)
    entry = pybibs.read_entry_string(bib)

    internals.unique_key_validation(entry["key"], data)

    data.append(entry)

    if file_:
        internals.set_file(data, entry, file_, destination, no_copy)

    pybibs.write_file(data, ctx.obj["database"])


@cli.command(short_help="Remove an entry or a field.")
@click.argument("key", shell_complete=internals.complete_key)
@click.argument("field", nargs=-1)
@click.pass_context
def remove(ctx, key, field):
    """
    Remove an entry from the database or remove a field from an entry.

    To remove an entry specify its key.
    To fields specify the key and list all fields for removal.
    """
    data = ctx.obj["data"]
    entry = query.get_by_key(data, key)

    if not field:
        data.remove(entry)
    elif "fields" in entry:
        for f in field:
            if f in entry["fields"]:
                del entry["fields"][f]
            else:
                click.echo('"{}" has no field "{}"'.format(key, f))
    else:
        click.echo('"{}" has no fields'.format(key))

    pybibs.write_file(data, ctx.obj["database"])


@cli.command(short_help="Edit an entry.")
@click.argument("key", shell_complete=internals.complete_key)
@click.argument("field_value", nargs=-1)
@FILE_OPTIONS
@click.pass_context
def edit(ctx, key, field_value, destination, no_copy, **kwargs):
    """
    Edit an entry.

    Use FIELD_VALUE to set fields as follows: ``author=Einstein``, or
    ``tags=interesting``.
    Leave the value empty to open in editor.
    Set the key or type in the same way.

    Don't forget to set the EDITOR environment variable for this command
    to work properly.
    """
    file_ = kwargs.pop("file")

    data = ctx.obj["data"]
    entry = query.get_by_key(data, key)

    if file_:
        internals.set_file(data, entry, file_, destination, no_copy)
    for fv in field_value:
        if "=" in fv:
            field, value = fv.split("=")
        else:
            field = fv
            current_value = entry["fields"].get(field, "")
            value = internals.editor(text=current_value).strip()
        if field == "key":
            internals.unique_key_validation(value, data)
            entry["key"] = value
        elif field == "type":
            entry["type"] = value
        else:
            entry["fields"][field] = value

    pybibs.write_file(data, ctx.obj["database"])


if __name__ == "__main__":
    cli()



import re


def split_entries(string):
    depth = 0
    start = 0
    for i, char in enumerate(string):
        if depth == 0 and char == "@":
            start = i
        if char == "{":
            depth += 1
        if char == "}":
            depth -= 1
            if depth == 0:
                yield string[start : i + 1]


def is_before_key(char, previous):
    return re.match(r"\s", char)


def is_key(char, previous):
    return char.isalpha()


def is_between(char, previous):
    together = "".join(previous + [char])
    if re.match(r"^\s*=?\s*$", together):
        return True


def is_value(char, previous):
    if previous and previous[0] == "{" and previous[-1] == "}":
        depth = 0
        for x in previous:
            if x == "{":
                depth += 1
            elif x == "}":
                depth -= 1
        if depth == 0:
            return False
    if len(previous) > 1 and previous[0] == previous[-1] == '"':
        return False
    if previous and previous[0] not in '{"' and char == ",":
        return False
    return True


def is_separator(char, previous):
    if char == ",":
        return True


def parse_raw_key_values(string):
    string = re.sub(r"\s+", " ", string)
    if not string.endswith(","):
        string += ","
    checks = [is_before_key, is_key, is_between, is_value, is_separator]
    content = []
    current = 0
    char_index = 0
    key = ""
    val = ""
    while char_index < len(string):
        char = string[char_index]
        if checks[current](char, content):
            content.append(char)
            char_index += 1
        else:
            if checks[current] == is_key:
                key = parse_key("".join(content))
            elif checks[current] == is_value:
                val = parse_value("".join(content))
                yield key, val
            content = []
            current = (current + 1) % len(checks)


def parse_key(key):
    return key.lower()


def parse_value(value):
    delimiters = [
        "{}",
        '""',
    ]
    value = value.strip()
    if value[-1] == ",":
        value = value[:-1]
    for delimiter in delimiters:
        if value[0] == delimiter[0] and value[-1] == delimiter[-1]:
            value = value[1:-1]
            break
    return value


def write_entry(entry):
    if entry["type"].lower() == "string":
        return write_string_entry(entry)
    elif entry["type"].lower() in ["comment", "preamble"]:
        return write_key_body_entry(entry)
    else:
        return write_general_entry(entry)


def write_string_entry(entry):
    return '@string{{{} = "{}"}}'.format(entry["key"], entry["val"])


def write_key_body_entry(entry):
    return "@{type}{{{body}}}".format_map(entry)


def write_general_entry(entry):
    parts = []
    parts += ["@", entry["type"], "{", entry["key"]]
    for k, v in entry["fields"].items():
        parts += [",\n", "  ", k, " = {", v, "}"]
    parts.append(",\n}")
    return "".join(parts)



import functools

import click


def constrain(constrained, depends=None, conflicts=None):
    """
    Constrain `constrained` to have all `depends` and non of `conflicts`.
    """

    if depends is None:
        depends = []
    if conflicts is None:
        conflicts = []

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if kwargs.get(constrained):
                for d in depends:
                    if not kwargs.get(d):
                        msg = "{} depends on {}".format(constrained, d)
                        raise click.BadOptionUsage(constrained, msg)
                for c in conflicts:
                    if kwargs.get(c):
                        msg = "{} conflicts with {}".format(constrained, c)
                        raise click.BadOptionUsage(constrained, msg)
            return f(*args, **kwargs)

        return wrapper

    return decorator



# coding=utf-8

import collections
import collections.abc
import importlib.metadata
import itertools
import os
import re
import shutil
import sys
import typing

import click

import pybibs

from . import models
from typing import Optional

BIBO_DATABASE_ENV_VAR = "BIBO_DATABASE"
_ANSI_BOLD = "\033[1m"
_ANSI_UNBOLD = "\033[22m"


def header(entry):
    parts = [click.style(entry["key"], fg="green")]
    fields = entry["fields"]
    if fields.get("tags"):
        parts.append(click.style(fields["tags"], fg="cyan"))
    if fields.get("file"):
        parts.append("📁")
    if fields.get("url") or fields.get("doi"):
        parts.append("🔗")
    return " ".join(parts)


def format_entry(entry, format_pattern):
    output = []
    replacement_start_index = -1
    for i, char in enumerate(format_pattern):
        if char == "$":
            replacement_start_index = i + 1
        elif (not char.isalpha()) and (replacement_start_index >= 0):
            field = format_pattern[replacement_start_index:i]
            output.append(_lookup(entry, field))
            output.append(char)
            replacement_start_index = -1
        elif replacement_start_index == -1:
            output.append(char)
    if replacement_start_index >= 0:
        field = format_pattern[replacement_start_index:]
        output.append(_lookup(entry, field))
    return "".join(output)


def _lookup(entry, field):
    if field in entry:
        return entry[field]
    if "fields" in entry and field in entry["fields"]:
        return entry["fields"][field]
    return "$" + field


def destination_heuristic(data):
    """
    A heuristic to get the folder with all other files from bib, using majority
    vote.
    """
    counter = collections.Counter()
    for entry in bib_entries(data):
        file_field = entry["fields"].get("file")
        if not file_field:
            continue
        path = os.path.dirname(file_field)
        counter[path] += 1

    if not counter:  # No paths found
        raise click.ClickException(
            "Path finding heuristics failed: no paths in the database"
        )

    # Find the paths that appears most often
    sorted_paths = counter.most_common()
    groupby = itertools.groupby(sorted_paths, key=lambda x: x[1])
    _, group = next(groupby)

    # We know that there's at least one candidate. Make sure it's
    # the only one
    candidate, _ = next(group)
    try:
        next(group)
    except StopIteration:
        return candidate
    else:
        raise click.ClickException(
            "Path finding heuristics failed: "
            "there are multiple equally valid paths in the database"
        )


def string_to_basename(s):
    """
    Converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    s = s.strip().lower()
    s = re.sub(r"[^\w\s-]", "", s)
    return re.sub(r"[\s-]+", "-", s)


def set_file(data, entry, file_, destination=None, no_copy=False):
    if no_copy:
        entry["fields"]["file"] = os.path.abspath(file_)
        return

    if not destination:
        destination = destination_heuristic(data)
    destination = os.path.abspath(destination)
    _, file_extension = os.path.splitext(file_)
    basename = string_to_basename(entry["key"])
    path = os.path.join(destination, basename + file_extension)
    entry["fields"]["file"] = path
    if os.path.exists(path):
        raise click.ClickException("{} already exists".format(path))
    click.echo("Copying {} to {}".format(file_, path))
    shutil.copy(file_, path)


def editor(*args, **kwargs):
    """
    Wrapper for `click.edit` that raises an error when None is returned.
    """
    result = click.edit(*args, **kwargs)
    if result is None:
        msg = "Editor exited without saving, command aborted"
        raise click.ClickException(msg)
    return result


def complete_key(ctx, param, incomplete):
    """
    Autocompletion for keys.
    """
    database = ctx.parent.params.get("database")
    if database:
        data = load_database(database)
    else:
        data = []
    data = bib_entries(data)
    return [x["key"] for x in data if x["key"].startswith(incomplete.lower())]


def load_database(database):
    """
    Load the database from .bib. Create (in memory) if doesn't exist.
    """
    try:
        return pybibs.read_file(database)
    except IOError:
        return []


def combine_decorators(decorators):
    # Copied from https://stackoverflow.com/a/4122845/1224456
    def decorator(f):
        for d in reversed(decorators):
            f = d(f)
        return f

    return decorator


def bib_entries(entries):
    """
    Yield only the actual bibliographic items from a list of entries.
    Drop @string / @comment / @preamble entries.
    """
    for e in entries:
        if e["type"].lower() not in ["string", "comment", "preamble"]:
            yield e


def unique_key_validation(new_key, data):
    if new_key in (e["key"] for e in bib_entries(data)):
        raise click.ClickException("Duplicate key, command aborted")


def bold(s: str) -> str:
    """Return `s` wrapped in ANSI bold."""
    return "{}{}{}".format(_ANSI_BOLD, s, _ANSI_UNBOLD)


def highlight_text(text: str, highlight: str) -> str:
    """
    Return `text` with sub-string `highlight` in bold.
    """
    # This is tricky because `text` might contain ANSI escape characters
    # that shouldn't get highlighted. A state machine to track whether
    # inside an ANSI sequence, and manual replacement with bold if found
    # otherwise.
    chars = list(text)
    n = len(highlight)

    in_ansi = False
    skip = 0
    res = []
    for i, char in enumerate(chars):
        if skip > 0:
            skip -= 1
            continue

        if not in_ansi and char == "\033":
            in_ansi = True

        candidate = "".join(chars[i : i + n])
        if not in_ansi and candidate.lower() == highlight.lower():
            res.append(bold(candidate))
            skip = n - 1
        else:
            res.append(char)

        if in_ansi and char == "m":
            in_ansi = False

    return "".join(res).replace(_ANSI_UNBOLD + _ANSI_BOLD, "")


def highlight_match(
    text: str, result: models.SearchResult, extra_match_info: Optional[dict] = None
) -> typing.Tuple[str, dict]:
    """
    Highlight `text` with `result.match` info. Every bit of info that is in
    `result.match` but not in `text` is added to the `extra_match_info` to
    present to the user.
    """
    if extra_match_info is None:
        extra_match_info = {}
    for key, vals in result.match.items():
        if isinstance(vals, collections.abc.Mapping):
            inner_result = models.SearchResult(
                result.entry[key],
                result.match[key],
            )
            text, extra_match_info = highlight_match(
                text,
                inner_result,
                extra_match_info,
            )
        else:
            for val in vals:
                if val.lower() in text.lower():
                    text = highlight_text(text, val)
                else:
                    extra_match_val = extra_match_info.get(key, result.entry[key])
                    extra_match_val = highlight_text(extra_match_val, val)
                    extra_match_info[key] = extra_match_val
    return text, extra_match_info


def get_plugins():
    """
    Return a list of EntryPoint objects for group "bibo.plugins".
    """
    group = "bibo.plugins"
    eps = importlib.metadata.entry_points()
    if sys.version_info >= (3, 12):
        return eps.select(group=group)
    else:
        return eps.get(group, [])



# Always prefer setuptools over distutils
from setuptools import setup, find_packages  # type: ignore

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="bibo",
    version="0.1.9",
    description="Command line reference manager with single source of truth: the .bib file. Inspired by beets",
    long_description=long_description,
    url="https://github.com/Nagasaki45/bibo",
    author="Tom Gurion",
    author_email="nagasaki45@gmail.com",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Database :: Front-Ends",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",
        "Topic :: Text Processing :: Markup :: LaTeX",
        "Topic :: Utilities",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords="bib latex reference interface",
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=["tests*"]),
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "click>=8",
        "click-plugins",
        "requests",
        "pylatexenc",
        "pyperclip",
    ],
    extras_require={
        "dev": [
            "black",
            "bumpversion",
            "coverage",
            "mypy",
            "pytest",
            "sphinx",
            "sphinx-click",
            "types-requests",
            "types-setuptools",
        ],
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={
        "console_scripts": [
            "bibo=bibo.bibo:cli",
        ],
    },
)



import collections
import collections.abc
import itertools
import re
import typing

import click

from . import internals, models


def search(data, search_terms: typing.Iterable[str]):
    if isinstance(search_terms, str):
        search_terms = [search_terms]
    search_terms = iter(search_terms)
    results = (models.SearchResult(e, {}) for e in internals.bib_entries(data))
    return _recursive_search(results, search_terms)


def _recursive_search(results, search_terms):
    try:
        search_term = next(search_terms)
        # Calculate match with search term and update results
        results = (_update_result(r, _match(r.entry, search_term)) for r in results)
        # Drop Nones with empty match
        results = (r for r in results if r)
        return _recursive_search(results, search_terms)
    except StopIteration:
        return results


def _match(entry, search_term: str):
    """
    Return a similar structure to an entry (nested dict) with matching strings
    as values.
    """
    # The match to populate
    d: typing.Dict[str, typing.Any] = {}

    # For cases where the entire search term is a key (e.g. best:author)
    _match_field("key", entry["key"], search_term, lambda: d)

    search_field, search_value = _parse_search_term(search_term)

    if search_field in ["key", "type"]:
        _match_field(search_field, entry[search_field], search_value, lambda: d)
    elif search_field in entry["fields"]:
        if search_value:
            _match_field(
                search_field,
                entry["fields"][search_field],
                search_value,
                lambda: d.setdefault("fields", {}),
            )
        # Allow query by field with no value (e.g. bibo list readdate:)
        else:
            d.setdefault("fields", {}).setdefault(search_field, set())
    elif search_field is None:
        for part in ["key", "type"]:
            _match_field(part, entry[part], search_value, lambda: d)
        for field, value in entry.get("fields", {}).items():
            _match_field(field, value, search_value, lambda: d.setdefault("fields", {}))
    return d


def _match_field(
    field: str,
    value: str,
    search_value: str,
    get_dict: typing.Callable[[], dict],
) -> None:
    """
    Try to match a field/value to a search_value. If there are
    matches, `get_dict` is called to get the dictionary to put the results
    in, usually the `match`, or `match["fields"]`.
    """
    matches = set(re.findall(search_value, value, re.IGNORECASE))
    matches.discard("")
    if matches:
        get_dict().setdefault(field, set()).update(matches)


def _update_result(result, new_match):
    if not new_match:
        return None
    d = _nested_update(result.match, new_match)
    return models.SearchResult(result.entry, d)


def _nested_update(d, u):
    _d = d.copy()
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            _d[k] = _nested_update(d.get(k, {}), v)
        else:
            _d[k] = _d.get(k, set())
            _d[k].update(v)
    return _d


def _parse_search_term(search_term: str):
    parts = search_term.split(":")
    if len(parts) == 1:
        return None, parts[0].lower()
    if len(parts) == 2:
        return parts[0].lower(), parts[1].lower()
    return "key", search_term.lower()


def get(data, search_terms):
    results = list(search(data, search_terms))

    for result in results:
        if result.entry["key"].lower() == search_terms[0].lower():
            return result

    if len(results) == 0:
        raise click.ClickException("No entries found")
    if len(results) > 1:
        raise click.ClickException("Multiple entries found")

    return results[0]


def get_by_key(data, key):
    for entry in internals.bib_entries(data):
        if entry["key"] == key:
            return entry
    raise click.ClickException('Couldn\'t find"{}"'.format(key))
